/**
 * Philm - class that handles all the simulations
 * 
 * @author Shing Hei Zhan, University of British Columbia 2010 iGEM team
 * @version 1.0 Oct. 27, 2010
 */
import java.lang.String.*;
import java.lang.Number.*;
import java.io.*;

public class Philm
{
    // constants
    private static final String newline = System.getProperty("line.separator");

    // starting conditions
    private double initialBiofilmPop;
    private double initialPhagePop;
    private double initialDepth;
    
    // parameters
    private Parameters pars;
    private History hist;
    private TotalBiofilmPop totBio;
    private InfectedBiofilmPop infBio;
    private UninfectedBiofilmPop uninfBio;
    private LatentBiofilmPop latBio;
    private LyticBiofilmPop lytBio;
    private PhagePop phage;
    private Diffusion diff;
    private AIP aip;
    private DspB dspb;
    private BiofilmStruct struct;
    
    /**
     * Constructor for objects of class Philm
     */
    public Philm(Parameters parSet)
    {
        pars = parSet;
        initialBiofilmPop = pars.get("iniBiof");
        initialPhagePop = pars.get("iniPhage");
        initialDepth = pars.get("iniDepth");
        totBio = new TotalBiofilmPop( initialBiofilmPop, pars.get("growthRateU"), pars.get("growthRateI"), pars.get("carryCap"), pars.get("lysisRate") );
        uninfBio = new UninfectedBiofilmPop( initialBiofilmPop, pars.get("growthRateU"), pars.get("carryCap") );
        infBio = new InfectedBiofilmPop( 0, pars.get("growthRateI"), pars.get("carryCap"), pars.get("lysisRate") );
        latBio = new LatentBiofilmPop( 0, pars.get("growthRateI"), pars.get("carryCap"), pars.get("lysisRate"), pars.get("convRate"), pars.get("portionLytic") );
        lytBio = new LyticBiofilmPop( 0, pars.get("growthRateI"), pars.get("carryCap"), pars.get("lysisRate"), pars.get("convRate"), pars.get("portionLytic") );
        phage = new PhagePop( initialPhagePop, pars.get("halfLife"), pars.get("infectRate"), pars.get("productRateR") );
        diff = new Diffusion( pars.get("rIni"), pars.get("rMin"), pars.get("rMax"), pars.get("rOut"), pars.get("Y") );
        aip = new AIP( pars.get("AIPmax"), pars.get("sigmaRate"), pars.get("gammaRate"), pars.get("AIPmax"), pars.get("AIPthres"), pars.get("carryCap") );
        dspb = new DspB( 0, pars.get("productRateS") );
        struct = new BiofilmStruct( initialDepth, 0, pars.get("occupancy"), pars.get("cellDiameter"), pars.get("crossArea") );
    }

    /**
     * modifier methods
     */
//    public static void main(String[] args){
//        Philm runIt = new Philm(args[0]);
//        History results = runIt.run((new Integer(args[2])).intValue());
//        runIt.printResults( results, args[1] );
//    }
    
    public History run(int genNum){
        double[] vars;
        hist = new History( genNum );    // zeroth row does not count as one generation
//        System.out.println("========= starting simulations =========");
        for( int x = 0; x < genNum; x++ ){
                                // 1 total biofilm pop. 2 uninfected subpop.    3 infected subpop.
            vars = new double[] {totBio.getPopCount(), uninfBio.getPopCount(), infBio.getPopCount(),
                // 4 latent subpop.   5 lytic subpop.       6 phage pop.         7 AIP         8 DspB
                latBio.getPopCount(), lytBio.getPopCount(), phage.getPopCount(), aip.getAIP(), dspb.getDspB(),
                // 9 rate       10 layer depth        11 biofilm depth   12 susceptibility
                diff.getRate(), diff.getPhageLayer(), struct.getDepth(), struct.getSuscept()};
             System.out.println("gen "+ x +": "+ vars[0] +" "+ vars[1] +" "+ vars[2] +" "+
               vars[3] +" "+ vars[4] +" "+ vars[5] +" "+ vars[6] +" "+ vars[7] +" "+
               vars[8] +" "+ vars[9] +" "+ vars[10] +" "+ vars[11]);
            // update populations NOTE: ORDER OF UPDATES MATTER; tweaking here may result in errors!!!
            uninfBio.updatePopCount(vars[0], vars[2], vars[5], vars[11], pars.get("infectRate"));
            infBio.updatePopCount(vars[0], vars[4], vars[5], vars[11], pars.get("infectRate"));
            latBio.updatePopCount(vars[0], vars[2], vars[5], vars[11], pars.get("infectRate"));
            lytBio.updatePopCount(vars[0], vars[2], vars[4], vars[3], vars[5], vars[11], pars.get("infectRate"));
            totBio.updatePopCount(uninfBio.getPopCount(), infBio.getPopCount());
            phage.updatePopCount(vars[0], vars[3], diff.getWashOffRate(), vars[11], pars.get("lysisRate"), vars[6], pars.get("AIPthres"));
            // update dissolved components
            dspb.updateDspB(diff.getWashOffRate(), vars[4], pars.get("lysisRate"), vars[6], pars.get("AIPthres"));
            aip.updateAIP(vars[0]);
            // update diffusion
            diff.updateRate(dspb.getDspB() - dspb.getDspBprev());
            diff.updatePhageLayer();
            diff.updateWashOffRate();
            // update structure
            struct.updateDepth( totBio.getPopCount() - totBio.getPopCountPrev() );
            struct.updateSuscept( diff.getPhageLayer() );
            // update records
            hist.updateHistory(vars, x);
            // stop simulations if all biofilm pop. gone
//            if( totBio.getPopCount() <= 1 ){
//                break;
//            }
        }
//        System.out.println("========= ending simulations =========");
        return hist;
    }
    
    public String printResults(History hist){
        try{
            String allOutput = "";
//            FileOutputStream fstream = new FileOutputStream( outFileName );
//            PrintStream printer = new PrintStream( fstream );
            double[] vars;
            double[] summ;
            // print header
//            printer.println("gen" +"	"+ "percBiofilmLeft" +"	"+ "percBiofilmInfected" +"	"+ "P-factor" +"	"+ "AIP" +"	"+ "DspB");
            for( int y = 0; y < hist.getColLen(); y++){
                vars = hist.getRowVal(y);
                //       1 % biofilm remaining                 2 % biofilm infected  3 P-factor
                double pfactor = Math.log(vars[5]/initialPhagePop);
                if(pfactor < -40){
                    pfactor = -40;
                }
                summ = new double[] { vars[0]/pars.get("carryCap"), vars[2]/vars[0], pfactor, vars[6], vars[7]};
//                printer.println(y +"	"+ vars[0] +"	"+ vars[1] +"	"+ vars[2] +"	"+
//                vars[3] +"	"+ vars[4] +"	"+ vars[5] +"	"+ vars[6] +"	"+ vars[7] +"	"+
//                vars[8] +"	"+ vars[9] +"	"+ vars[10] +"	"+ vars[11]);
//                printer.println(y +"	"+ summ[0] +"	"+ summ[1] +"	"+ summ[2] +"	"+ summ[3] +"	"+ summ[4]);
                allOutput = allOutput + y +"	"+ summ[0] +"	"+ summ[1] +"	"+ summ[2] +"	"+ summ[3] +"	"+ summ[4] + newline;
            }
//            printer.close();
            return allOutput;
        }catch(Exception e){
            System.out.println("failed to create output file...");
            return "some error occurred...";
        }
    }
}